from itertools import chain

import numpy as np
    
from Orange.data import Table


def find_lat_lon(data):
    """Return inferred latitude and longitude attributes as found in the data domain"""
    assert isinstance(data, Table)

    all_vars = list(chain(data.domain.variables, data.domain.metas))

    lat_attr = next(
        (attr for attr in all_vars
         if attr.is_continuous and
         attr.name.lower().startswith(('latitude', 'lat'))), None)
    lon_attr = next(
        (attr for attr in all_vars
         if attr.is_continuous and
         attr.name.lower().startswith(('longitude', 'lng', 'long', 'lon'))),
        None)

    def _all_between(vals, min, max):
        return np.all((min <= vals) & (vals <= max))

    if not lat_attr:
        for attr in all_vars:
            if attr.is_continuous:
                values = np.nan_to_num(
                    data.get_column_view(attr)[0].astype(float))
                if _all_between(values, -90, 90):
                    lat_attr = attr
                    break
    if not lon_attr:
        for attr in all_vars:
            if attr.is_continuous:
                values = np.nan_to_num(
                    data.get_column_view(attr)[0].astype(float))
                if _all_between(values, -180, 180):
                    lon_attr = attr
                    break

    return lat_attr, lon_attr



# Category metadata.

# Category icon show in the menu
ICON = "icons/category.svg"

# Background color for category background in menu
# and widget icon background in workflow.
BACKGROUND = "#a0eaa0"

# Location of widget help files.
WIDGET_HELP_PATH = (
# Used for development.
# You still need to build help pages using
# make htmlhelp
# inside doc folder
("{DEVELOP_ROOT}/doc/build/htmlhelp/index.html", None),

# Online documentation url, used when the local documentation is available.
# Url should point to a page with a section Widgets. This section should
# includes links to documentation pages of each widget. Matching is
# performed by comparing link caption to widget name.
("http://orange3-geo.readthedocs.io/en/latest/", "")
)



import re
from itertools import chain


_CC_NAME_TO_CC_NAME = {
    # EU
    'Kosovo': ['Kosovo, Republic of'],
    'Northern Cyprus': ['N. Cyprus', 'North Cyprus'],
    'Aland': ['Åland Islands', 'Åland'],
    'Bosnia and Herzegovina': ['Bosnia and Herz.'],
    'Czech Republic': ['Czech Rep.'],
    'Faroe Islands': ['Faeroe Is.'],
    'United Kingdom': ['Great Britain', 'England'],
    'Moldova': ['Moldova, Republic of'],
    'Macedonia': ['Macedonia, Republic of'],
    'Palestine': ['Palestine, State of'],
    'Russia': ['Russian Federation'],
    'Syria': ['Syrian Arab Republic'],
    # World
    'Brunei': ['Brunei Darussalam'],
    'Bolivia': ['Bolivia, Plurinational State of'],
    'Dem. Rep. Congo': ['Democratic Republic of the Congo', 'Democratic Republic of Congo', 'Congo, The Democratic Republic of the'],
    'Central African Rep.': ['Central African Republic'],
    'Dominican Republic': ['Dominican Rep.'],
    'W. Sahara': ['Western Sahara'],
    'Falkland Is.': ['Falkland Islands [Malvinas]', 'Falkland Islands', 'Malvinas'],
    'Eq. Guinea': ['Equatorial Guinea'],
    'Iran': ['Iran, Islamic Republic of'],
    'Dem. Rep. Korea': ["Korea, Democratic People's Republic of", 'Democratic Republic of Korea', 'North Korea'],
    'South Korea': ['Korea, Republic of', 'Korea'],
    'Lao PDR': ["Lao People's Democratic Republic", "Laos"],
    'Solomon Is.': ['Solomon Islands'],
    'S. Sudan': ['South Sudan'],
    'Taiwan': ['Taiwan, Province of China'],
    'Tanzania': ['Tanzania, United Republic of'],
    'United States': ['United States of America'],
    'Venezuela': ['Venezuela, Bolivarian Republic of'],
    'Vietnam': ['Viet Nam'],
}


_REGION_NAME_TO_REGION_NAME = {
    # Municipalities in Slovenia
    'Apace': ['Apače'],
    'Braslovce': ['Braslovče'],
    'Crnomelj': ['Črnomelj'],
    'Crnomelj': ['Črnomelj'],
    'Divaca': ['Divača'],
    'Hoce-Slivnica': ['Hoče-Slivnica'],
    'Ivancna Gorica': ['Ivančna Gorica'],
    'Kidricevo': ['Kidričevo'],
    'Kocevje': ['Kočevje'],
    'Luce': ['Luče'],
    'Mirna Pec': ['Mirna Peč'],
    'Moravce': ['Moravče'],
    'Nova Goriška': ['Nova Gorica'],
    'Novo Mesto': ['Novo mesto'],
    'Podcetrtek': ['Podčetrtek'],
    'Poljcane': ['Poljčane'],
    'Race-Fram': ['Rače-Fram'],
    'Radece': ['Radeče'],
    'Ravne na Koroškem': ['Ravne'],
    'Semic': ['Semič'],
    'Šmartno in Litiji': ['Šmartno pri Litiji'],
    'Solcava': ['Solčava'],
    'Sveti Andraž v Slovenskih Goricah': ['Sveti Andraž v Slovenskih goricah', 'Sveti Andraž v Slov. goricah', 'Sveti Andraž v Slov.goricah', 'Sveti Andraž v Sloven. goricah'],
    'Zasavska': ['Zagorje ob Savi'],
    'Zavrc': ['Zavrč'],
    'Zrece': ['Zreče'],
}


_US_STATE_TO_US_STATE = {
    'NC': ['North Carolina', 'N. Carolina'],
    'ND': ['North Dakota', 'N. Dakota'],
    'SC': ['South Carolina', 'S. Carolina'],
    'SD': ['South Dakota', 'S. Dakota'],
}


_US_CITIES = {
    # From https://en.wikipedia.org/wiki/List_of_United_States_cities_by_population
    'AK': ['Anchorage'],
    'AL': ['Birmingham', 'Montgomery', 'Mobile', 'Huntsville'],
    'AR': ['Little Rock'],
    'AZ': ['Phoenix', 'Tucson', 'Mesa', 'Chandler', 'Gilbert', 'Glendale', 'Scottsdale', 'Tempe', 'Peoria', 'Surprise'],
    'CA': ['Los Angeles', 'San Diego', 'San Jose', 'San Francisco', 'Fresno', 'Sacramento', 'Long Beach', 'Oakland', 'Bakersfield', 'Anaheim', 'Santa Ana', 'Riverside', 'Stockton', 'Chula Vista', 'Irvine', 'Fremont', 'San Bernardino', 'Modesto', 'Oxnard', 'Fontana', 'Moreno Valley', 'Huntington Beach', 'Glendale', 'Santa Clarita', 'Garden Grove', 'Oceanside', 'Rancho Cucamonga', 'Santa Rosa', 'Ontario', 'Elk Grove', 'Corona', 'Lancaster', 'Palmdale', 'Salinas', 'Hayward', 'Pomona', 'Escondido', 'Sunnyvale', 'Torrance', 'Pasadena', 'Orange', 'Fullerton', 'Thousand Oaks', 'Visalia', 'Roseville', 'Concord', 'Simi Valley', 'Santa Clara', 'Victorville', 'Vallejo', 'Berkeley', 'El Monte', 'Downey', 'Costa Mesa', 'Carlsbad', 'Inglewood', 'Fairfield', 'Ventura', 'Temecula', 'Antioch', 'Richmond', 'West Covina', 'Murrieta', 'Norwalk', 'Daly City', 'Burbank', 'Santa Maria', 'El Cajon', 'San Mateo', 'Rialto', 'Clovis', 'East Los Angeles'],
    'CO': ['Denver', 'Colorado Springs', 'Aurora', 'Fort Collins', 'Lakewood', 'Thornton', 'Arvada', 'Westminster', 'Pueblo', 'Centennial', 'Boulder'],
    'CT': ['Bridgeport', 'New Haven', 'Stamford', 'Hartford', 'Waterbury'],
    'DE': ['Dover',],
    'DC': ['Washington', 'Washington, D.C.', 'Washington D.C.', 'Washington DC', 'D.C.'],
    'FL': ['Jacksonville', 'Miami', 'Tampa', 'Orlando', 'St. Petersburg', 'Hialeah', 'Tallahassee', 'Fort Lauderdale', 'Port St. Lucie', 'Cape Coral', 'Pembroke Pines', 'Hollywood', 'Miramar', 'Gainesville', 'Coral Springs', 'Miami Gardens', 'Clearwater', 'Pompano Beach', 'Palm Bay', 'West Palm Beach', 'Lakeland', 'Brandon'],
    'GA': ['Atlanta', 'Columbus', 'Augusta', 'Macon', 'Savannah', 'Athens', 'Sandy Springs'],
    'HI': ['Honolulu'],
    'IA': ['Des Moines', 'Cedar Rapids', 'Davenport'],
    'ID': ['Boise'],
    'IL': ['Chicago', 'Aurora', 'Rockford', 'Joliet', 'Naperville', 'Springfield', 'Peoria', 'Elgin'],
    'IN': ['Indianapolis', 'Fort Wayne', 'Evansville', 'South Bend'],
    'KS': ['Wichita', 'Overland Park', 'Kansas City', 'Olathe', 'Topeka'],
    'KY': ['Louisville', 'Lexington'],
    'LA': ['New Orleans', 'Baton Rouge', 'Shreveport', 'Lafayette', 'Metairie'],
    'MA': ['Boston', 'Worcester', 'Springfield', 'Lowell', 'Cambridge'],
    'MD': ['Annapolis', 'Baltimore'],
    'ME': ['Portland', 'Augusta'],
    'MI': ['Detroit', 'Grand Rapids', 'Warren', 'Sterling Heights', 'Ann Arbor', 'Lansing'],
    'MN': ['Minneapolis', 'St. Paul', 'Rochester'],
    'MO': ['Kansas City', 'St. Louis', 'Springfield', 'Independence', 'Columbia'],
    'MS': ['Jackson'],
    'MT': ['Billings'],
    'NC': ['Charlotte', 'Raleigh', 'Greensboro', 'Durham', 'Winston–Salem', 'Fayetteville', 'Cary', 'Wilmington', 'High Point'],
    'ND': ['Fargo'],
    'NE': ['Omaha', 'Lincoln'],
    'NH': ['Manchester'],
    'NJ': ['Newark', 'Jersey City', 'Paterson', 'Elizabeth'],
    'NM': ['Albuquerque', 'Las Cruces'],
    'NV': ['Las Vegas', 'Henderson', 'Reno', 'North Las Vegas', 'Paradise', 'Sunrise Manor', 'Spring Valley', 'Enterprise'],
    'NY': ['New York City', 'Buffalo', 'Rochester', 'Yonkers', 'Syracuse'],
    'OH': ['Columbus', 'Cleveland', 'Cincinnati', 'Toledo', 'Akron', 'Dayton'],
    'OK': ['Oklahoma City', 'Tulsa', 'Norman', 'Broken Arrow'],
    'OR': ['Portland', 'Salem', 'Eugene', 'Gresham'],
    'PA': ['Philadelphia', 'Pittsburgh', 'Allentown'],
    'RI': ['Providence'],
    'SC': ['Columbia', 'Charleston', 'North Charleston'],
    'SD': ['Sioux Falls'],
    'TN': ['Memphis', 'Nashville', 'Knoxville', 'Chattanooga', 'Clarksville', 'Murfreesboro'],
    'TX': ['Houston', 'San Antonio', 'Dallas', 'Austin', 'Fort Worth', 'El Paso', 'Arlington', 'Corpus Christi', 'Plano', 'Laredo', 'Lubbock', 'Garland', 'Irving', 'Amarillo', 'Grand Prairie', 'Brownsville', 'McKinney', 'Pasadena', 'Frisco', 'Mesquite', 'McAllen', 'Killeen', 'Waco', 'Carrollton', 'Denton', 'Midland', 'Abilene', 'Beaumont', 'Odessa', 'Round Rock', 'Richardson', 'Wichita Falls', 'College Station', 'Pearland', 'Lewisville', 'Tyler'],
    'UT': ['Salt Lake City', 'West Valley City', 'Provo', 'West Jordan'],
    'VA': ['Virginia Beach', 'Norfolk', 'Chesapeake', 'Richmond', 'Newport News', 'Alexandria', 'Hampton', 'Arlington County'],
    'VT': ['Burlington'],
    'WA': ['Seattle', 'Spokane', 'Tacoma', 'Vancouver', 'Bellevue', 'Kent', 'Everett'],
    'WI': ['Milwaukee', 'Madison', 'Green Bay'],
    'WV': ['Charleston'],
    'WY': ['Cheyenne'],
}

_EUROPE_CITIES = {
    # As by https://www.countries-ofthe-world.com/capitals-of-europe.html
    'PS': ['Pristina', 'Priština', 'Priština', 'Prishtina', 'Prishtinë'],
    'AD': ['Andorra la Vella'],
    'AL': ['Tirana'],
    'AT': ['Vienna', 'Wien'],
    'AX': ['Mariehamn'],
    'BA': ['Sarajevo'],
    'BE': ['Brussels', 'Bruxelles', 'Brussel'],
    'BG': ['Sofia'],
    'BY': ['Minsk'],
    'CH': ['Bern', 'Zürich', 'Zurich', 'Geneva'],
    'CY': ['Nicosia', 'North Nicosia'],
    'CZ': ['Prague', 'Praha', 'Praga'],
    'DE': ['Berlin', 'Hanover', 'Hamburg', 'Munich', 'München', 'Stuttgart', 'Frankfurt', 'Düsseldorf', 'Cologne', 'Essen', 'Dresden', 'Leipzig'],
    'DK': ['Copenhagen', 'København'],
    'DZ': ['Algiers'],
    'EE': ['Tallinn'],
    'EG': ['Cairo'],
    'ES': ['Madrid', 'Barcelona', 'Murcia', 'Malaga', 'Seville', 'Bilbao', 'Zaragoza', 'Valencia'],
    'FI': ['Helsinki'],
    'FO': ['Tórshavn', 'Torshavn'],
    'FR': ['Paris', 'Nice', 'Montpellier', 'Marseille', 'Lyon', 'Toulouse', 'Bordeaux', 'Nantes'],
    'GB': ['London', 'London Heathrow', 'Manchester', 'Liverpool', 'Bristol', 'Cardiff', 'Oxford', 'Cambridge', 'Glasgow', 'Edinburgh', 'Belfast'],
    'GE': ['Tbilisi'],
    'GG': ['Saint Peter Port'],
    'GR': ['Athens', 'Thessaloniki'],
    'HR': ['Zagreb'],
    'HU': ['Budapest'],
    'IE': ['Dublin'],
    'IL': ['Jerusalem', 'Tel Aviv'],
    'IM': ['Douglas'],
    'IQ': ['Baghdad'],
    'IS': ['Reykjavik'],
    'IT': ['Rome', 'Roma'],
    'JE': ['Saint Helier'],
    'JO': ['Amman'],
    'LB': ['Beirut'],
    'LI': ['Vaduz'],
    'LT': ['Vilnius'],
    'LU': ['Luxembourg'],
    'LV': ['Riga'],
    'LY': ['Tripoli'],
    'MA': ['Rabat', 'Fes', 'Casablanca', 'Marrakesh', 'Marrakech', 'Tangier'],
    'MD': ['Chisinau', 'Chișinău'],
    'ME': ['Podgorica'],
    'MK': ['Skopje'],
    'MT': ['Valletta'],
    'NL': ['Amsterdam', 'The Hague', 'Rotterdam'],
    'NO': ['Oslo', 'Bergen', 'Trondheim'],
    'PL': ['Warsaw', 'Warszawa', 'Krakow', 'Lublin', 'Bialystok', 'Lodz', 'Gdansk', 'Wroclaw'],
    'PT': ['Lisbon', 'Porto'],
    'RO': ['Bucharest', 'București'],
    'RS': ['Belgrade', 'Beograd'],
    'RU': ['Moscow', 'Moskva', 'Saint Petersburg', 'St. Petersburg', 'Kazan', 'Vladivostok', 'Irkutsk', 'Novosibirsk', 'Krasnoyarsk', 'Nizhny Novgorod'],
    'SA': ['Riyadh'],
    'SE': ['Stockholm', 'Gothenburg', 'Malmo'],
    'SI': ['Ljubljana'],
    'SK': ['Bratislava'],
    'SM': ['San Marino'],
    'SY': ['Damascus'],
    'TN': ['Tunis'],
    'TR': ['Ankara', 'Istanbul', 'Constantinople'],
    'UA': ['Kyiv', 'Kiev', 'Kharkiv', 'Odessa', 'Lviv'],
}


_WORLD_CITIES = {
    # Capitals and larger cities. Does NOT include CC_EUROPE.
    'AE': ['Abu Dhabi'],
    'AF': ['Kabul'],
    'AM': ['Yerevan'],
    'AO': ['Luanda'],
    'AR': ['Buenos Aires'],
    'AU': ['Canberra', 'Sydney', 'Melbourne', 'Cairns', 'Adelaide', 'Perth'],
    'AZ': ['Baku'],
    'BD': ['Dhaka'],
    'BF': ['Ouagadougou'],
    'BI': ['Bujumbura'],
    'BJ': ['Porto-Novo', 'Porto Novo', 'Cotonou'],
    'BN': ['Bandar Seri Begawani'],
    'BO': ['La Paz', 'Sucre'],
    'BR': ['Brasília', 'Brasilia', 'Sao Paulo', 'Sao Paulo', 'Rio de Janeiro'],
    'BS': ['Nassau'],
    'BT': ['Thimphu'],
    'BW': ['Gaborone'],
    'BZ': ['Belmopan'],
    'CA': ['Ottawa', 'Vancouver', 'Calgary', 'Edmonton', 'Winnipeg', 'Toronto', 'Montreal', 'Québec', 'Quebec'],
    'CD': ['Kinshasa'],
    'CF': ['Bangui'],
    'CG': ['Brazzaville'],
    'CI': ['Yamoussoukro', 'Abidjan'],
    'CL': ['Santiago'],
    'CM': ['Yaoundé', 'Yaounde'],
    'CN': ['Beijing'],
    'CO': ['Bogotá', 'Bogota'],
    'CR': ['San José', 'San Jose'],
    'CU': ['Havana'],
    'DJ': ['Djibouti'],
    'DO': ['Santo Domingo'],
    'EC': ['Quito'],
    'EH': ['Laayoune'],
    'ER': ['Asmara'],
    'ET': ['Addis Ababa'],
    'FJ': ['Suva'],
    'FK': ['Stanley'],
    'GA': ['Libreville'],
    'GH': ['Accra'],
    'GL': ['Nuuk'],
    'GM': ['Banjul'],
    'GN': ['Conakry'],
    'GQ': ['Malabo'],
    'GT': ['Guatemala City'],
    'GW': ['Bissau'],
    'GY': ['Georgetown'],
    'HN': ['Tegucigalpa'],
    'HT': ['Port-au-Prince', 'Port au Prince'],
    'ID': ['Jakarta'],
    'IN': ['New Delhi', 'Mumbai', 'Bombay', 'Bengaluru', 'Hyderabad'],
    'IR': ['Tehran'],
    'JM': ['Kingston'],
    'JP': ['Tokyo', 'Osaka', 'Hiroshima', 'Fukuoka', 'Sapporo'],
    'KE': ['Nairobi'],
    'KG': ['Bishkek'],
    'KH': ['Phnom Penh'],
    'KP': ['Pyongyang'],
    'KR': ['Seoul'],
    'KW': ['Kuwait City'],
    'KZ': ['Astana'],
    'LA': ['Vientiane'],
    'LK': ['Sri Jayawardenepura Kotte', 'Colombo'],
    'LR': ['Monrovia'],
    'LS': ['Maseru'],
    'MG': ['Antananarivo'],
    'ML': ['Bamako'],
    'MM': ['Naypyidaw'],
    'MN': ['Ulaanbaatar', 'Ulan Bator'],
    'MR': ['Nouakchott'],
    'MW': ['Lilongwe'],
    'MX': ['Mexico City'],
    'MY': ['Kuala Lumpur', 'Putrajaya'],
    'MZ': ['Maputo'],
    'NA': ['Windhoek'],
    'NC': ['Nouméa', 'Noumea'],
    'NE': ['Niamey'],
    'NG': ['Abuja'],
    'NI': ['Managua'],
    'NP': ['Kathmandu'],
    'NZ': ['Wellington', 'Auckland', 'Christchurch', 'Queenstown'],
    'OM': ['Muscat'],
    'PA': ['Panama City'],
    'PE': ['Lima'],
    'PG': ['Port Moresby'],
    'PH': ['Manila'],
    'PK': ['Islamabad'],
    'PR': ['San Juan'],
    'PY': ['Asunción', 'Asuncion'],
    'QA': ['Doha'],
    'RW': ['Kigali'],
    'SB': ['Honiara'],
    'SD': ['Khartoum'],
    'SL': ['Freetown'],
    'SN': ['Dakar'],
    'SO': ['Mogadishu'],
    'SR': ['Paramaribo'],
    'SS': ['Juba'],
    'SV': ['San Salvador'],
    'SZ': ['Mbabane', 'Lobamba'],
    'TD': ['N\'Djamena', 'NDjamena', 'N Djamena'],
    'TG': ['Lomé', 'Lome'],
    'TH': ['Bangkok'],
    'TJ': ['Dushanbe'],
    'TL': ['Dili'],
    'TM': ['Ashgabat'],
    'TT': ['Port of Spain'],
    'TW': ['Taipei'],
    'TZ': ['Dodoma'],
    'UG': ['Kampala'],
    'UY': ['Montevideo'],
    'UZ': ['Tashkent'],
    'VE': ['Caracas'],
    'VN': ['Hanoi', 'Ho Chi Minh'],
    'VU': ['Port Vila'],
    'YE': ['Sana\'a', 'Sanaa'],
    'ZA': ['Pretoria', 'Bloemfontein', 'Cape Town', 'Johannesburg'],
    'ZM': ['Lusaka'],
    'ZW': ['Harare'],
}

assert not (set(_WORLD_CITIES) & set(_EUROPE_CITIES))
assert 'US' not in _WORLD_CITIES
assert 'US' not in _EUROPE_CITIES


def _invert(mapping):
    return {v.lower(): k
            for k in mapping
            for v in mapping[k]}


def _regexify(mapping):
    """Make key literals a "contains key" regex so the mapping can be used
    with pandas' .replace()"""
    return {re.compile(r'.*\b' + k + r'\b.*', re.IGNORECASE): v
            for k, v in mapping.items()}


CC_NAME_TO_CC_NAME = _regexify(_invert(_CC_NAME_TO_CC_NAME))
REGION_NAME_TO_REGION_NAME = _regexify(_invert(_REGION_NAME_TO_REGION_NAME))
US_STATE_TO_US_STATE = _regexify(_invert(_US_STATE_TO_US_STATE))

EUROPE_CITIES = _regexify(_invert(_EUROPE_CITIES))
US_CITIES = _regexify(_invert(_US_CITIES))

WORLD_CITIES = _WORLD_CITIES.copy()
WORLD_CITIES.update(US=list(chain.from_iterable(_US_CITIES.values())))
WORLD_CITIES.update(_EUROPE_CITIES)
WORLD_CITIES = _regexify(_invert(WORLD_CITIES))

EUROPE_CITIES_LIST = sorted(chain.from_iterable(_EUROPE_CITIES.values()))
US_CITIES_LIST = sorted(chain.from_iterable(_US_CITIES.values()))
WORLD_CITIES_LIST = sorted(set(chain(chain.from_iterable(_WORLD_CITIES.values()),
                                     EUROPE_CITIES_LIST,
                                     US_CITIES_LIST)))



from unittest import TestCase

import numpy as np

from orangecontrib.geo.mapper import latlon2region, get_bounding_rect


class TestUtils(TestCase):

    def test_mapper(self):
        latlons = np.array([
            [46.0555, 14.5083],       # Ljubljana
            [40.7127, -74.0059],      # NYC
            [32.775833, -96.796667],  # Dallas
            (49.761667, -77.802778),  # Quebec, Ontario
            (31.814700, 79.886838),   # Tibet
            (52.818925, 92.567674),   # Krasnoyarsk
            [64.295556, -15.227222],  # Austurland, Iceland
            [61.760153, -121.236525], # Canada
            [0, -1],                  # "middle of the ocean"
            [40, np.nan],
        ])
        self.assertEqual(
            [i.get('name') for i in latlon2region(latlons, 2)],
            ['Ljubljana', 'New York, New York', 'Dallas, Texas', 'Québec',
             'Xizang', 'Krasnoyarsk', 'Austurland', 'Northwest Territories',
             None, None])
        self.assertEqual(
            [i.get('_id') for i in latlon2region(latlons, 0)],
             ['SVN', 'USA', 'USA', 'CAN', 'CHN', 'RUS', 'ISL', 'CAN', None, None])
        self.assertEqual(
            [i.get('_id') for i in latlon2region(latlons, 1)],
            ['SVN-962', 'USA-3559', 'USA-3536', 'CAN-683', 'CHN-1662',
             'RUS-2603', 'ISL-695', 'CAN-635', None, None])
        self.assertEqual(
            [i.get('_id') for i in latlon2region(latlons, 2)],
            ['SVN-962', 'USA-NY-36061', 'USA-TX-48113', 'CAN-683', 'CHN-1662',
             'RUS-2603', 'ISL-695', 'CAN-635', None, None])

    def test_get_bounding_rect(self):
        np.testing.assert_equal(np.array(get_bounding_rect({'SWE', 'ITA'}), dtype=int),
                                np.r_[12, 42, 14, 62])



import os
import logging

import numpy as np
import pandas as pd
from scipy import stats

from AnyQt.QtCore import (
    Qt, QUrl, pyqtSignal, pyqtSlot, QT_VERSION,
    QObject, QTimer,
)

from Orange.misc.cache import memoize_method
from Orange.util import color_to_hex
from Orange.data import Table, TimeVariable, DiscreteVariable, ContinuousVariable
from Orange.widgets import gui, widget, settings
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.utils.webview import WebviewWidget
from Orange.widgets.utils.annotated_data import create_annotated_table, ANNOTATED_DATA_SIGNAL_NAME
from Orange.widgets.widget import Input, Output

from orangecontrib.geo.utils import find_lat_lon
from orangecontrib.geo.mapper import latlon2region, ADMIN2_COUNTRIES, get_bounding_rect


if QT_VERSION <= 0x050300:
    raise RuntimeError('Choropleth widget only works with Qt 5.3+')


log = logging.getLogger(__name__)


# Test that memoize_method exposes cache_clear(), avoid otherwise
# https://github.com/biolab/orange3/pull/2229
if not hasattr(memoize_method(1)(int), 'cache_clear'):
    memoize_method = lambda x: (lambda x: x)


class LeafletChoropleth(WebviewWidget):
    selectionChanged = pyqtSignal(list)

    def __init__(self, parent=None):

        class Bridge(QObject):
            @pyqtSlot()
            def fit_to_bounds(_):
                return self.fit_to_bounds()

            @pyqtSlot('QVariantList')
            def selection(_, selected):
                self.selectionChanged.emit(selected)

        super().__init__(parent,
                         bridge=Bridge(),
                         url=QUrl(self.toFileURL(
                             os.path.join(os.path.dirname(__file__), '_leaflet', 'owchoropleth.html'))))
        self._owwidget = parent
        self.bounds = None

    def fit_to_bounds(self):
        if self.bounds is None:
            return
        east, south, west, north = self.bounds
        maxzoom = 5 if self._owwidget.admin == 0 else 7
        self.evalJS('''
            map.flyToBounds([[%f, %f], [%f, %f]], {
                padding: [0,0], minZoom: 2, maxZoom: %d,
                duration: .6, easeLinearity: .4
            });''' % (north, west, south, east, maxzoom))

    def set_opacity(self, opacity):
        self.evalJS('''set_opacity(%f);''' % (opacity / 100))

    def set_quantization(self, quantization):
        self.evalJS('''set_quantization("%s");''' % (quantization[0].lower()))

    def set_color_steps(self, steps):
        self.evalJS('''set_color_steps(%d);''' % steps)

    def toggle_legend(self, visible):
        self.evalJS('''toggle_legend(%d);''' % (int(bool(visible))))

    def toggle_map_labels(self, visible):
        self.evalJS('''toggle_map_labels(%d);''' % (int(bool(visible))))

    def toggle_tooltip_details(self, visible):
        self.evalJS('''toggle_tooltip_details(%d);''' % (int(bool(visible))))

    def preset_region_selection(self, selection):
        self.evalJS('''set_region_selection(%s);''' % selection)


class OWChoropleth(widget.OWWidget):
    name = 'Choropleth Map'
    description = 'A thematic map in which areas are shaded in proportion ' \
                  'to the measurement of the statistical variable being displayed.'
    icon = "icons/Choropleth.svg"
    priority = 120

    class Inputs:
        data = Input("Data", Table, default=True)

    class Outputs:
        selected_data = Output("Selected Data", Table, default=True)
        annotated_data = Output(ANNOTATED_DATA_SIGNAL_NAME, Table)

    settingsHandler = settings.DomainContextHandler()

    want_main_area = True

    AGG_FUNCS = (
        'Count',
        'Count defined',
        'Sum',
        'Mean',
        'Median',
        'Mode',
        'Max',
        'Min',
        'Std',
    )
    AGG_FUNCS_TRANSFORM = {
        'Count': 'size',
        'Count defined': 'count',
        'Mode': lambda x: stats.mode(x, nan_policy='omit').mode[0],
    }
    AGG_FUNCS_DISCRETE = ('Count', 'Count defined', 'Mode')
    AGG_FUNCS_CANT_TIME = ('Count', 'Count defined', 'Sum', 'Std')

    autocommit = settings.Setting(True)
    lat_attr = settings.ContextSetting('')
    lon_attr = settings.ContextSetting('')
    attr = settings.ContextSetting('')
    agg_func = settings.ContextSetting(AGG_FUNCS[0])
    admin = settings.Setting(0)
    opacity = settings.Setting(70)
    color_steps = settings.Setting(5)
    color_quantization = settings.Setting('equidistant')
    show_labels = settings.Setting(True)
    show_legend = settings.Setting(True)
    show_details = settings.Setting(True)
    selection = settings.ContextSetting([])

    class Error(widget.OWWidget.Error):
        aggregation_discrete = widget.Msg("Only certain types of aggregation defined on categorical attributes: {}")

    class Warning(widget.OWWidget.Warning):
        logarithmic_nonpositive = widget.Msg("Logarithmic quantization requires all values > 0. Using 'equidistant' quantization instead.")

    graph_name = "map"

    def __init__(self):
        super().__init__()
        self.map = map = LeafletChoropleth(self)
        self.mainArea.layout().addWidget(map)
        self.selection = []
        self.data = None
        self.latlon = None
        self.result_min_nonpositive = False
        self._should_fit_bounds = False

        def selectionChanged(selection):
            self._indices = self.ids.isin(selection).nonzero()[0]
            self.selection = selection
            self.commit()

        map.selectionChanged.connect(selectionChanged)

        box = gui.vBox(self.controlArea, 'Aggregation')

        self._latlon_model = DomainModel(parent=self, valid_types=ContinuousVariable)
        self._combo_lat = combo = gui.comboBox(
            box, self, 'lat_attr', orientation=Qt.Horizontal,
            label='Latitude:', sendSelectedValue=True, callback=self.aggregate)
        combo.setModel(self._latlon_model)

        self._combo_lon = combo = gui.comboBox(
            box, self, 'lon_attr', orientation=Qt.Horizontal,
            label='Longitude:', sendSelectedValue=True, callback=self.aggregate)
        combo.setModel(self._latlon_model)

        self._combo_attr = combo = gui.comboBox(
            box, self, 'attr', orientation=Qt.Horizontal,
            label='Attribute:', sendSelectedValue=True, callback=self.aggregate)
        combo.setModel(DomainModel(parent=self, valid_types=(ContinuousVariable, DiscreteVariable)))

        gui.comboBox(
            box, self, 'agg_func', orientation=Qt.Horizontal, items=self.AGG_FUNCS,
            label='Aggregation:', sendSelectedValue=True, callback=self.aggregate)

        self._detail_slider = gui.hSlider(
            box, self, 'admin', None, 0, 2, 1,
            label='Administrative level:', labelFormat=' %d',
            callback=self.aggregate)

        box = gui.vBox(self.controlArea, 'Visualization')

        gui.spin(box, self, 'color_steps', 3, 15, 1, label='Color steps:',
                 callback=lambda: self.map.set_color_steps(self.color_steps))

        def _set_quantization():
            self.Warning.logarithmic_nonpositive(
                shown=(self.color_quantization.startswith('log') and
                       self.result_min_nonpositive))
            self.map.set_quantization(self.color_quantization)

        gui.comboBox(box, self, 'color_quantization', label='Color quantization:',
                     orientation=Qt.Horizontal, sendSelectedValue=True,
                     items=('equidistant', 'logarithmic', 'quantile', 'k-means'),
                     callback=_set_quantization)

        self._opacity_slider = gui.hSlider(
            box, self, 'opacity', None, 20, 100, 5,
            label='Opacity:', labelFormat=' %d%%',
            callback=lambda: self.map.set_opacity(self.opacity))

        gui.checkBox(box, self, 'show_legend', label='Show legend',
                     callback=lambda: self.map.toggle_legend(self.show_legend))
        gui.checkBox(box, self, 'show_labels', label='Show map labels',
                     callback=lambda: self.map.toggle_map_labels(self.show_labels))
        gui.checkBox(box, self, 'show_details', label='Show region details in tooltip',
                     callback=lambda: self.map.toggle_tooltip_details(self.show_details))

        gui.rubber(self.controlArea)
        gui.auto_commit(self.controlArea, self, 'autocommit', 'Send Selection')

        self.map.toggle_legend(self.show_legend)
        self.map.toggle_map_labels(self.show_labels)
        self.map.toggle_tooltip_details(self.show_details)
        self.map.set_quantization(self.color_quantization)
        self.map.set_color_steps(self.color_steps)
        self.map.set_opacity(self.opacity)

    def __del__(self):
        self.map = None

    def commit(self):
        if self.data is not None and self.selection:
            selected_data = self.data[self._indices]
            annotated_data = create_annotated_table(self.data, self._indices)
        else:
            selected_data = annotated_data = None

        self.Outputs.selected_data.send(selected_data)
        self.Outputs.annotated_data.send(annotated_data)

    @Inputs.data
    def set_data(self, data):
        self.data = data

        self.closeContext()

        self.clear()
        
        if data is None:
            return

        self._combo_attr.model().set_domain(data.domain)
        self._latlon_model.set_domain(data.domain)

        lat, lon = find_lat_lon(data)
        if lat or lon:
            self._combo_lat.setCurrentIndex(-1 if lat is None else self._latlon_model.indexOf(lat))
            self._combo_lon.setCurrentIndex(-1 if lat is None else self._latlon_model.indexOf(lon))
            self.lat_attr = lat.name if lat else None
            self.lon_attr = lon.name if lon else None
            if lat and lon:
                self.latlon = np.c_[self.data.get_column_view(self.lat_attr)[0],
                                    self.data.get_column_view(self.lon_attr)[0]]

        if data.domain.class_var:
            self.attr = data.domain.class_var.name
        else:
            self.attr = self._combo_attr.itemText(0)

        self.openContext(data)

        if self.selection:
            self.map.preset_region_selection(self.selection)
        self.aggregate()

        self.map.set_opacity(self.opacity)

        if self.isVisible():
            self.map.fit_to_bounds()
        else:
            self._should_fit_bounds = True

    def showEvent(self, event):
        super().showEvent(event)
        if self._should_fit_bounds:
            QTimer.singleShot(500, self.map.fit_to_bounds)
            self._should_fit_bounds = False

    def aggregate(self):
        if self.latlon is None or self.attr not in self.data.domain:
            self.clear(caches=False)
            return

        attr = self.data.domain[self.attr]

        if attr.is_discrete and self.agg_func not in self.AGG_FUNCS_DISCRETE:
            self.Error.aggregation_discrete(', '.join(map(str.lower, self.AGG_FUNCS_DISCRETE)))
            self.Warning.logarithmic_nonpositive.clear()
            self.clear(caches=False)
            return
        else:
            self.Error.aggregation_discrete.clear()

        try:
            regions, adm0, result, self.map.bounds = \
                self.get_grouped(self.lat_attr, self.lon_attr, self.admin, self.attr, self.agg_func)
        except ValueError:
            # This might happen if widget scheme File→Choropleth, and
            # some attr is selected in choropleth, and then the same attr
            # is set to string attr in File and dataset reloaded.
            # Our "dataflow" arch can suck my balls
            return

        # Only show discrete values that are contained in aggregated results
        discrete_values = []
        if attr.is_discrete and not self.agg_func.startswith('Count'):
            subset = sorted(result.drop_duplicates().dropna().astype(int))
            discrete_values = np.array(attr.values)[subset].tolist()
            discrete_colors = np.array(attr.colors)[subset].tolist()
            result = result.replace(dict(zip(subset, range(len(subset)))))

        self.result_min_nonpositive = attr.is_continuous and result.min() <= 0
        force_quantization = self.color_quantization.startswith('log') and self.result_min_nonpositive
        self.Warning.logarithmic_nonpositive(shown=force_quantization)

        repr_time = isinstance(attr, TimeVariable) and self.agg_func not in self.AGG_FUNCS_CANT_TIME

        self.map.exposeObject(
            'results',
            dict(discrete=discrete_values,
                 colors=[color_to_hex(i)
                         for i in (discrete_colors if discrete_values else
                                   ((0, 0, 255), (255, 255, 0)) if attr.is_discrete else
                                   attr.colors[:-1])],  # ???
                 regions=list(adm0),
                 attr=attr.name,
                 have_nonpositive=self.result_min_nonpositive or bool(discrete_values),
                 values=result.to_dict(),
                 repr_vals=result.map(attr.repr_val).to_dict() if repr_time else {},
                 minmax=([result.min(), result.max()] if attr.is_discrete and not discrete_values else
                         [attr.repr_val(result.min()), attr.repr_val(result.max())] if repr_time or not discrete_values else
                         [])))

        self.map.evalJS('replot();')

    @memoize_method(3)
    def get_regions(self, lat_attr, lon_attr, admin):
        latlon = np.c_[self.data.get_column_view(lat_attr)[0],
                       self.data.get_column_view(lon_attr)[0]]
        regions = latlon2region(latlon, admin)
        adm0 = ({'0'} if admin == 0 else
                {'1-' + a3 for a3 in (i.get('adm0_a3') for i in regions) if a3} if admin == 1 else
                {('2-' if a3 in ADMIN2_COUNTRIES else '1-') + a3
                 for a3 in (i.get('adm0_a3') for i in regions) if a3})
        ids = [i.get('_id') for i in regions]
        self.ids = pd.Series(ids)
        regions = set(ids) - {None}
        bounds = get_bounding_rect(regions) if regions else None
        return regions, ids, adm0, bounds

    @memoize_method(6)
    def get_grouped(self, lat_attr, lon_attr, admin, attr, agg_func):
        log.debug('Grouping %s(%s) by (%s, %s; admin%d)',
                  agg_func, attr, lat_attr,  lon_attr, admin)
        regions, ids, adm0, bounds = self.get_regions(lat_attr, lon_attr, admin)
        attr = self.data.domain[attr]
        result = pd.Series(self.data.get_column_view(attr)[0], dtype=float)\
            .groupby(ids)\
            .agg(self.AGG_FUNCS_TRANSFORM.get(agg_func, agg_func.lower()))
        return regions, adm0, result, bounds

    def clear(self, caches=True):
        if caches:
            try:
                self.get_regions.cache_clear()
                self.get_grouped.cache_clear()
            except AttributeError:
                pass  # back-compat https://github.com/biolab/orange3/pull/2229
        self.selection = []
        self.map.exposeObject('results', {})
        self.map.evalJS('replot();')


def main():
    from AnyQt.QtWidgets import QApplication
    a = QApplication([])

    ow = OWChoropleth()
    ow.show()
    ow.raise_()
    data = Table("India_census_district_population")
    ow.set_data(data)

    a.exec()
    ow.saveSettings()

if __name__ == "__main__":
    main()



import time
from threading import Thread
from functools import lru_cache, wraps

from operator import itemgetter
from os import path
from glob import glob
import logging

import simplejson as json
import numpy as np
import pandas as pd
from sklearn.neighbors import KDTree

import shapely.speedups
from shapely.geometry import Point, shape as Shape, Polygon

from orangecontrib.geo.cc_cities import \
    CC_NAME_TO_CC_NAME, REGION_NAME_TO_REGION_NAME, US_STATE_TO_US_STATE,\
    EUROPE_CITIES, US_CITIES, WORLD_CITIES,\
    EUROPE_CITIES_LIST, US_CITIES_LIST, WORLD_CITIES_LIST



log = logging.getLogger(__name__)


def is_shapely_speedups_available():
    if not shapely.speedups.available:
        return False
    # Otherwise try shapely with speedups in a subprocess to see if shit
    # is crash due to ABI-incompatible libgeos found in path on Loonix
    import sys, subprocess
    proc = subprocess.Popen(
        sys.executable + ' -c ' + '''"import json
from shapely.geometry import shape
import shapely.speedups
shapely.speedups.enable()
shape(json.load(open('%s'))['features'][0]['geometry'])
"''' % path.join(GEOJSON_DIR, 'admin0.json'),
        # Didn't return correct exit status without shell=True
        shell=True)
    proc.wait()
    return proc.returncode == 0


GEOJSON_DIR = path.join(path.dirname(__file__), 'geojson')

ADMIN2_COUNTRIES = {path.basename(filename).split('.')[0].split('-')[1]
                    for filename in glob(path.join(GEOJSON_DIR, 'admin2-*.json'))}
NUL = {}  # nonmapped (invalid) output region


if is_shapely_speedups_available():
    shapely.speedups.enable()
    log.debug('Shapely speed-ups available')


def wait_until_loaded(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        global is_done_loading
        while not is_done_loading:
            time.sleep(.1)
        if isinstance(is_done_loading, Exception):
            raise is_done_loading
        return func(*args, **kwargs)
    return wrapper


is_done_loading = False


def init():
    global is_done_loading
    def _admin_cc(filename):
        parts = path.basename(filename).split('.', 1)[0].split('-')
        admin, cc = parts if len(parts) == 2 else (parts[0], None)
        admin = int(admin[-1])
        return admin, cc

    nearest_points = {0: [], 1: [], 2: []}
    shapes = {0: [], 1: [], 2: []}
    cc_shapes = {}
    id_regions = {}
    us_states = {}

    files = glob(path.join(GEOJSON_DIR, 'admin*.json'))
    if not files:
        is_done_loading = RuntimeError(
            'Missing GeoJSON files. '
            'In development environments, merge in the "json"'
            'branch. See CONTRIBUTING.md')
        raise is_done_loading

    log.debug('Loading GeoJSON data ...')
    for filename in files:
        admin, cc = _admin_cc(filename)

        with open(filename, encoding='utf-8') as f:
            collection = json.load(f, encoding='utf-8')

        for feature in collection['features']:
            p = feature['properties']
            shape = Shape(feature['geometry'])
            tup = (shape, p)

            # Add representative lat-lon pair if non-existent
            if (np.isnan(p.get('latitude', np.nan)) or
                np.isnan(p.get('longitude', np.nan))):
                try:
                    r = shape.representative_point()
                except ValueError:
                    # For GBR, representative point above fails with:
                    #   ValueError: No Shapely geometry can be created from null value
                    r = shape.centroid
                    if not r.within(shape):
                        max_poly = max([shape] if isinstance(shape, Polygon) else list(shape),
                                       key=lambda polygon: polygon.area)
                        # From https://stackoverflow.com/questions/33311616/find-coordinate-of-closest-point-on-polygon-shapely/33324058#33324058
                        poly_ext = max_poly.exterior
                        dist = poly_ext.project(r)
                        pt = poly_ext.interpolate(dist)
                        r = Point(pt.coords[0])
                assert r.within(shape)
                p.update(latitude=r.y, longitude=r.x)

            # Get representative points for the k-d tree
            points = []
            polygons = (shape,) if isinstance(shape, Polygon) else shape
            for poly in polygons:
                points.append([poly.centroid.y, poly.centroid.x])
                if poly.area > 10:
                    points.extend(xy[::-1]
                                  for xy in poly.simplify(1).exterior.coords)

            nearest_points[admin].extend(points)
            tups = [tup] * len(points)
            shapes[admin].extend(tups)
            id_regions[p['_id']] = tup

            if admin == 0:
                cc_shapes[p['adm0_a3']] = tup

            # Make Admin1 shapes available in Admin2 too. Except for USA
            # which is the country we have explicit Admin2 shapes for
            if admin == 1 and cc not in ADMIN2_COUNTRIES:
                shapes[2].extend(tups)
                nearest_points[2].extend(points)

            if admin == 1 and cc == 'USA':
                us_states[p['hasc'].split('.')[1]] = tup

    kdtree = {admin: KDTree(centroids)
              for admin, centroids in nearest_points.items()}
    cc_shapes['NUL'] = (None, NUL)  # tuple for Null Island

    assert all(len(nearest_points[admin]) == len(shapes[admin])
               for admin in shapes)

    global SHAPES, CC_SHAPES, KDTREE, ID_REGIONS, US_STATES
    SHAPES, CC_SHAPES, KDTREE, ID_REGIONS, US_STATES = shapes, cc_shapes, kdtree, id_regions, us_states
    is_done_loading = True
    return

Thread(target=init).start()


class ToLatLon:
    @classmethod
    def _lookup(cls, mapping, key):
        return {p[key]: p
                for _, p in mapping.values()
                if key in p}

    @classmethod
    def _get(cls, lookup, values, to_replace={}, _NUL={}):
        mapping = values.drop_duplicates()
        mapping.index = mapping.values.copy()
        mapping.replace(to_replace, inplace=True)
        mapping = mapping.apply(lookup.get, args=(_NUL,))
        return values.map(mapping).tolist()

    @classmethod
    @wait_until_loaded
    def from_cc_name(cls, values):
        return cls._get(ToLatLon._lookup(CC_SHAPES, 'name'), values, CC_NAME_TO_CC_NAME)

    @classmethod
    @wait_until_loaded
    def from_cc2(cls, values):
        return cls._get(ToLatLon._lookup(CC_SHAPES, 'iso_a2'), values)

    @classmethod
    @wait_until_loaded
    def from_cc3(cls, values):
        return cls._get(ToLatLon._lookup(CC_SHAPES, 'iso_a3'), values)

    @classmethod
    @wait_until_loaded
    def from_region(cls, values):
        return cls._get(ToLatLon._lookup(ID_REGIONS, 'name'), values, REGION_NAME_TO_REGION_NAME)

    @classmethod
    @wait_until_loaded
    def from_fips(cls, values):
        return cls._get(ToLatLon._lookup(ID_REGIONS, 'fips'), values)

    @classmethod
    @wait_until_loaded
    def from_hasc(cls, values):
        return cls._get(ToLatLon._lookup(ID_REGIONS, 'hasc'), values)

    @classmethod
    @wait_until_loaded
    def from_us_state(cls, values):
        lookup = ToLatLon._lookup(US_STATES, 'name')
        lookup.update({short_name: data
                       for short_name, (_polygon, data) in US_STATES.items()})
        return cls._get(lookup, values, US_STATE_TO_US_STATE)

    @classmethod
    @wait_until_loaded
    def from_city_eu(cls, values):
        assert isinstance(values, pd.Series)
        return cls.from_cc2(values.replace(regex=EUROPE_CITIES))

    @classmethod
    @wait_until_loaded
    def from_city_us(cls, values):
        assert isinstance(values, pd.Series)
        return cls.from_us_state(values.replace(regex=US_CITIES))

    @classmethod
    @wait_until_loaded
    def from_city_world(cls, values):
        assert isinstance(values, pd.Series)
        return cls.from_cc2(values.replace(regex=WORLD_CITIES))

    @classmethod
    @lru_cache(1)
    @wait_until_loaded
    def valid_values(cls, method):
        """ Return a sorted list of valid values for method of ToLatLon """
        assert method.__name__.startswith('from_')
        lookup_args = {
            ToLatLon.from_region: (ID_REGIONS, 'name'),
            ToLatLon.from_cc2: (CC_SHAPES, 'iso_a2'),
            ToLatLon.from_cc3: (CC_SHAPES, 'iso_a3'),
            ToLatLon.from_cc_name: (CC_SHAPES, 'name'),
            ToLatLon.from_fips: (ID_REGIONS, 'fips'),
            ToLatLon.from_hasc: (ID_REGIONS, 'hasc'),
            ToLatLon.from_us_state: (US_STATES, 'name'),
            ToLatLon.from_us_state: (US_STATES, 'name'),
            ToLatLon.from_city_eu: (EUROPE_CITIES_LIST,),
            ToLatLon.from_city_us: (US_CITIES_LIST,),
            ToLatLon.from_city_world: (WORLD_CITIES_LIST,),
        }[method]
        if len(lookup_args) == 1:
            return lookup_args[0]
        return sorted(filter(None, ToLatLon._lookup(*lookup_args).keys()))

    @classmethod
    def detect_input(cls, values, sample_size=200):
        """
        Return first "from_" method that in more than 50% matches values,
        or None.
        """
        assert isinstance(values, pd.Series)
        values = values.drop_duplicates().dropna()
        if len(values) > sample_size:
            values = values.sample(sample_size)
        strlen = values.str.len().dropna().unique()
        for method, *cond in ((cls.from_cc2, len(strlen) == 1 and strlen[0] == 2),
                              (cls.from_cc3, len(strlen) == 1 and strlen[0] == 3),
                              (cls.from_cc_name,),
                              (cls.from_us_state,),
                              (cls.from_city_eu,),
                              (cls.from_city_us,),
                              (cls.from_city_world,),
                              (cls.from_region,),
                              (cls.from_fips,),
                              (cls.from_hasc, np.in1d(strlen, [2, 5, 8]).all())):
            if cond and not cond[0]:
                continue
            if sum(map(bool, method(values))) >= len(values) / 2:
                return method
        return None


@wait_until_loaded
def latlon2region(latlon, admin=0):
    """Return list of property dicts for regions mapped by latlon coordinates"""
    assert len(latlon) == 0 or len(latlon[0]) == 2
    assert 0 <= admin <= 2
    global SHAPES, CC_SHAPES, KDTREE

    latlon = np.asanyarray(latlon, dtype=float)

    log.debug('Mapping %d coordinate pairs into regions', len(latlon))

    # Replace missing latlon data with invalid coordinates for k-d tree to work
    nan_rows = np.isnan(latlon).any(axis=1)
    if nan_rows.any():
        latlon = np.array(latlon, order='C', copy=True)
        latlon[nan_rows, :] = -500

    @lru_cache(700)
    def resolve_coords(coord):
        """Cached resolution.
        shape.contains(point) test is what takes the most time
        """
        nonlocal inds, shapes
        point = Point(*coord)
        for i in inds:
            shape, props = shapes[i]
            if shape.contains(point):
                return props
        # No shapes contain point. See if distance to nearest neighbor
        # is less than threshold (i.e. point very near but outside shape)
        shape, props = shapes[sorted(inds, key=lambda i: shapes[i][0].distance(point))[0]]
        return props if shape.distance(point) < .2 else NUL

    out = []
    shapes = SHAPES[admin or 1]
    for isnan, coord, inds in zip(nan_rows,
                                  np.roll(latlon, -1, axis=1),
                                  KDTREE[admin or 1].query(latlon, k=30,
                                                           return_distance=False,
                                                           sort_results=True)):
        out.append(NUL if isnan else resolve_coords(tuple(coord)))

    if admin == 0:
        out = [i and CC_SHAPES[i['adm0_a3']][1] for i in out]

    assert len(out) == len(latlon)
    return out


@wait_until_loaded
def get_bounding_rect(region_ids):
    """Return lat-lon bounding rect of the union of regions defined by ids"""
    if not region_ids:
        return None
    coords = itemgetter('longitude', 'latitude')
    centroids = np.array([coords(ID_REGIONS[_id][1])
                          for _id in region_ids])
    mins, maxs = centroids.min(0), centroids.max(0)
    return tuple(mins.tolist() + maxs.tolist())


if __name__ == '__main__':
    from pprint import pprint

    pprint(ToLatLon.from_cc2(pd.Series(['RU', 'SI'])))

    coords = [
        (46.0555, 14.5083),
        (40.7127, -74.0059),
        (49.761667, -77.802778),
        (31.814700, 79.886838),
        (52.818925, 92.567674),
        (61.760153, -121.236525),
        (64.295556,-15.227222),
        (0, 0),
    ]

    start = time.clock()
    pprint(latlon2region(coords, 0))
    elpassed = time.clock() - start
    print(elpassed)
    print()

    start = time.clock()
    pprint(latlon2region(coords, 1))
    elpassed = time.clock() - start
    print(elpassed)
    print()

    start = time.clock()
    pprint(latlon2region(coords, 2))
    elpassed = time.clock() - start
    print(elpassed)



import logging

import numpy as np
import pandas as pd
from collections import OrderedDict

from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import QComboBox, QItemEditorFactory, QLineEdit, QCompleter

from Orange.data import Table, Domain, StringVariable, DiscreteVariable, ContinuousVariable
from Orange.widgets import gui, widget, settings
from Orange.widgets.utils.itemmodels import DomainModel, PyTableModel
from Orange.widgets.widget import Input, Output
from orangecontrib.geo.utils import find_lat_lon
from orangecontrib.geo.mapper import latlon2region, ToLatLon


log = logging.getLogger(__name__)


def available_name(domain, template):
    """Return the next available variable name (from template) that is not
    already taken in domain"""
    for i in range(100000):
        name = '{}{}'.format(template, ('_' + str(i)) if i else '')
        if name not in domain:
            return name


class OWGeocoding(widget.OWWidget):
    name = 'Geocoding'
    description = 'Encode region names into geographical coordinates, or ' \
                  'reverse-geocode latitude and longitude pairs into cultural ' \
                  'regions.'
    icon = "icons/Geocoding.svg"
    priority = 40

    class Inputs:
        data = Input("Data", Table, default=True)

    class Outputs:
        coded_data = Output("Coded Data", Table, default=True)

    settingsHandler = settings.DomainContextHandler()

    resizing_enabled = False

    ID_TYPE = OrderedDict((
        ('Country name', ToLatLon.from_cc_name),
        ('ISO 3166-1 alpha-2 country code', ToLatLon.from_cc2),
        ('ISO 3166-1 alpha-3 country code', ToLatLon.from_cc3),
        # ('ISO 3166-1 numeric-3 country code', ),  # Who uses this?
        # ('ISO 3166-2 subdivision code', ),  # XXX: where to get these?
        ('Region name', ToLatLon.from_region),
        ('Major city (US)', ToLatLon.from_city_us),
        ('Major city (Europe)', ToLatLon.from_city_eu),
        ('Major city (World)', ToLatLon.from_city_world),
        ('FIPS code', ToLatLon.from_fips),
        ('HASC code', ToLatLon.from_hasc),
        ('US state (name or abbr.)', ToLatLon.from_us_state),
    ))

    autocommit = settings.Setting(True)
    is_decoding = settings.ContextSetting(1)
    str_attr = settings.ContextSetting('')
    str_type = settings.ContextSetting(next(iter(ID_TYPE)))
    lat_attr = settings.ContextSetting('')
    lon_attr = settings.ContextSetting('')
    admin = settings.ContextSetting(0)
    append_features = settings.Setting(True)

    replacements = settings.Setting([], schema_only=True)

    class Error(widget.OWWidget.Error):
        aggregation_discrete = widget.Msg("Only certain types of aggregation defined on categorical attributes: {}")

    class Warning(widget.OWWidget.Warning):
        logarithmic_nonpositive = widget.Msg("Logarithmic quantization requires all values > 0. Using 'equidistant' quantization instead.")

    def __init__(self):
        super().__init__()
        self.data = None
        self.domainmodels = []

        top = self.controlArea

        def _radioChanged():
            self.mainArea.setVisible(self.is_decoding == 0)
            self.commit()

        modes = gui.radioButtons(top, self, 'is_decoding', callback=_radioChanged)

        gui.appendRadioButton(
            modes, '&Encode region names into geographical coordinates:', insertInto=top)
        box = gui.indentedBox(top)
        model = DomainModel(parent=self, valid_types=(StringVariable, DiscreteVariable))
        self.domainmodels.append(model)

        def _region_attr_changed():
            if self.data is None:
                return

            # Auto-detect the type of region in the attribute and set its combo
            values = self._get_data_values()
            func = ToLatLon.detect_input(values)
            str_type = next((k for k, v in self.ID_TYPE.items() if v == func), None)
            if str_type is not None and str_type != self.str_type:
                self.str_type = str_type

            self.commit()

        combo = gui.comboBox(
            box, self, 'str_attr', label='Region identifier:', orientation=Qt.Horizontal,
            callback=_region_attr_changed, sendSelectedValue=True)
        combo.setModel(model)
        gui.comboBox(
            box, self, 'str_type', label='Identifier type:', orientation=Qt.Horizontal,
            items=tuple(self.ID_TYPE.keys()), callback=lambda: self.commit(), sendSelectedValue=True)

        # Select first mode if any of its combos are changed
        for combo in box.findChildren(QComboBox):
            combo.currentIndexChanged.connect(
                lambda: setattr(self, 'is_decoding', 0))

        gui.appendRadioButton(
            modes, '&Decode latitude and longitude into regions:', insertInto=top)
        box = gui.indentedBox(top)
        model = DomainModel(parent=self, valid_types=ContinuousVariable)
        self.domainmodels.append(model)
        combo = gui.comboBox(
            box, self, 'lat_attr', label='Latitude:', orientation=Qt.Horizontal,
            callback=lambda: self.commit(), sendSelectedValue=True)
        combo.setModel(model)
        combo = gui.comboBox(
            box, self, 'lon_attr', label='Longitude:', orientation=Qt.Horizontal,
            callback=lambda: self.commit(), sendSelectedValue=True)
        combo.setModel(model)
        gui.comboBox(
            box, self, 'admin', label='Administrative level:', orientation=Qt.Horizontal,
            callback=lambda: self.commit(),
            items=('Country',
                   '1st-level subdivision (state, region, province, municipality, ...)',
                   '2nd-level subdivisions (1st-level & US counties)'),)

        # Select second mode if any of its combos are changed
        for combo in box.findChildren(QComboBox):
            combo.currentIndexChanged.connect(
                lambda: setattr(self, 'is_decoding', 1))

        gui.checkBox(
            top, self, 'append_features',
            label='E&xtend coded data with additional region properties',
            callback=lambda: self.commit(),
            toolTip='Extend coded data with region properties, such as'
                    'ISO codes, continent, subregion, region type, '
                    'economy type, FIPS/HASC codes, region capital etc. as available.')

        gui.auto_commit(self.controlArea, self, 'autocommit', '&Apply')

        model = self.replacementsModel = PyTableModel(self.replacements, parent=self, editable=[False, True])
        view = gui.TableView(self,
                             sortingEnabled=False,
                             selectionMode=gui.TableView.NoSelection,
                             editTriggers=gui.TableView.AllEditTriggers)
        view.horizontalHeader().setSectionResizeMode(0)
        view.verticalHeader().setSectionResizeMode(0)
        view.setModel(model)

        owwidget = self

        class EditorFactory(QItemEditorFactory):
            def createEditor(self, p_int, parent):
                nonlocal owwidget
                edit = QLineEdit(parent)
                wordlist = [''] + ToLatLon.valid_values(owwidget.ID_TYPE[owwidget.str_type])
                edit.setCompleter(
                    QCompleter(wordlist, edit,
                               caseSensitivity=Qt.CaseInsensitive,
                               filterMode=Qt.MatchContains))
                return edit

        self.factory = EditorFactory()
        view.itemDelegate().setItemEditorFactory(self.factory)
        model.setHorizontalHeaderLabels(['Unmatched Identifier', 'Custom Replacement'])
        box = gui.vBox(self.mainArea)
        self.info_str = ' /'
        gui.label(box, self, 'Unmatched identifiers: %(info_str)s')
        box.layout().addWidget(view)
        self.mainArea.setVisible(self.is_decoding == 0)

    def commit(self):
        output = None
        if self.data is not None and len(self.data):
            output = self.decode() if self.is_decoding else self.encode()
            if output is not None:
                output = self.data.concatenate((self.data, output))
        self.Outputs.coded_data.send(output)

    def decode(self):
        if (self.data is None or not len(self.data) or
                self.lat_attr not in self.data.domain or
                self.lon_attr not in self.data.domain):
            return None
        latlon = np.c_[self.data.get_column_view(self.lat_attr)[0],
                       self.data.get_column_view(self.lon_attr)[0]]
        assert isinstance(self.admin, int)
        with self.progressBar(2) as progress:
            progress.advance()
            regions = pd.DataFrame(latlon2region(latlon, self.admin))
        return self._to_addendum(regions, ['name'])

    def encode(self):
        if self.data is None or not len(self.data) or self.str_attr not in self.data.domain:
            return None
        values = self._get_data_values()
        log.debug('Geocoding %d regions into coordinates', len(values))
        with self.progressBar(4) as progress:
            progress.advance()
            mappings = self.ID_TYPE[self.str_type](values)

            progress.advance()
            invalid_idx = [i for i, value in enumerate(mappings) if not value]
            unmatched = values[invalid_idx].drop_duplicates().dropna().sort_values()
            self.info_str = '{} / {}'.format(len(unmatched), values.nunique())

            replacements = {k: v
                            for k, v in self.replacementsModel.tolist()
                            if v}
            self.replacements = ([[name, '']
                                  for name in unmatched
                                  if name not in replacements] +
                                 [[name, value]
                                  for name, value in replacements.items()])
            self.replacementsModel.wrap(self.replacements)

            progress.advance()
            latlon = pd.DataFrame(mappings)
        return self._to_addendum(latlon, ['latitude', 'longitude'])

    def _get_data_values(self):
        if self.data is None:
            return None
        values = self.data.get_column_view(self.str_attr)[0]
        # no comment
        if self.data.domain[self.str_attr].is_discrete:
            values = np.array(self.data.domain[self.str_attr].values)[values.astype(np.int16)].astype(str)
        values = pd.Series(values)

        # Apply replacements from the replacements table
        if len(self.replacementsModel):
            values = values.replace({k: v
                                     for k, v in self.replacementsModel.tolist()
                                     if v})
        return values

    def _to_addendum(self, df, keep):
        if not df.shape[1]:
            return None

        df.drop(['_id', 'adm0_a3'], axis=1, inplace=True)
        addendum = df if self.append_features else df[keep]
        table = Table(Domain(
            [], metas=[(ContinuousVariable if col in ('latitude', 'longitude') else
                        StringVariable)(available_name(self.data.domain, col))
                       for col in addendum]),
            np.empty((len(addendum), 0)), None, addendum.values)
        return table

    @Inputs.data
    def set_data(self, data):
        self.data = data
        self.closeContext()

        if data is None or not len(data):
            self.commit()
            return

        for model in self.domainmodels:
            model.set_domain(data.domain)

        lat, lon = find_lat_lon(data)
        self.lat_attr = lat.name if lat else None
        self.lon_attr = lon.name if lon else None

        self.openContext(data)
        self.mainArea.setVisible(self.is_decoding == 0)
        self.commit()


def main():
    from AnyQt.QtWidgets import QApplication
    a = QApplication([])

    ow = OWGeocoding()
    ow.show()
    ow.raise_()
    data = Table("India_census_district_population")
    print(data[:10])
    ow.set_data(data)

    a.exec()
    ow.saveSettings()

if __name__ == "__main__":
    main()



import os
from itertools import chain, repeat
from collections import OrderedDict
from tempfile import mkstemp

import numpy as np

from AnyQt.QtCore import Qt, QUrl, pyqtSignal, pyqtSlot, QTimer, QT_VERSION, QObject
from AnyQt.QtGui import QImage, QPainter, QPen, QBrush, QColor
from AnyQt.QtWidgets import qApp


from Orange.util import color_to_hex
from Orange.base import Learner
from Orange.data.util import scale
from Orange.data import Table, Domain, TimeVariable, DiscreteVariable, ContinuousVariable
from Orange.widgets import gui, widget, settings
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.utils.webview import WebviewWidget
from Orange.widgets.utils.colorpalette import ColorPaletteGenerator, ContinuousPaletteGenerator
from Orange.widgets.utils.annotated_data import create_annotated_table, ANNOTATED_DATA_SIGNAL_NAME
from Orange.widgets.widget import Input, Output

from orangecontrib.geo.utils import find_lat_lon


if QT_VERSION <= 0x050300:
    raise RuntimeError('Map widget only works with Qt 5.3+')


class LeafletMap(WebviewWidget):
    selectionChanged = pyqtSignal(list)

    def __init__(self, parent=None):

        class Bridge(QObject):
            @pyqtSlot()
            def fit_to_bounds(_):
                return self.fit_to_bounds()

            @pyqtSlot(float, float, float, float)
            def selected_area(_, *args):
                return self.selected_area(*args)

            @pyqtSlot('QVariantList')
            def recompute_heatmap(_, *args):
                return self.recompute_heatmap(*args)

            @pyqtSlot(float, float, float, float, int, int, float, 'QVariantList', 'QVariantList')
            def redraw_markers_overlay_image(_, *args):
                return self.redraw_markers_overlay_image(*args)

        super().__init__(parent,
                         bridge=Bridge(),
                         url=QUrl(self.toFileURL(
                             os.path.join(os.path.dirname(__file__), '_leaflet', 'owmap.html'))),
                         debug=True,)
        self.jittering = 0
        self._jittering_offsets = None
        self._owwidget = parent
        self._opacity = 255
        self._sizes = None
        self._selected_indices = None

        self.lat_attr = None
        self.lon_attr = None
        self.data = None
        self.model = None
        self._domain = None
        self._latlon_data = None

        self._jittering = None
        self._color_attr = None
        self._label_attr = None
        self._shape_attr = None
        self._size_attr = None
        self._legend_colors = []
        self._legend_shapes = []
        self._legend_sizes = []

        self._drawing_args = None
        self._image_token = None
        self._prev_map_pane_pos = None
        self._prev_origin = None
        self._overlay_image_path = mkstemp(prefix='orange-Map-', suffix='.png')[1]
        self._subset_ids = np.array([])
        self.is_js_path = None

        self._should_fit_bounds = False

    def __del__(self):
        os.remove(self._overlay_image_path)
        self._image_token = np.nan

    def set_data(self, data, lat_attr, lon_attr):
        self.data = data
        self._image_token = np.nan  # Stop drawing previous image
        self._owwidget.progressBarFinished(None)
        self._owwidget.Warning.all_nan_slice.clear()

        if (data is None or not len(data) or
                lat_attr not in data.domain or
                lon_attr not in data.domain):
            self.data = None
            self.evalJS('clear_markers_js(); clear_markers_overlay_image();')
            self._legend_colors = []
            self._legend_shapes = []
            self._legend_sizes = []
            self._update_legend()
            return

        lat_attr = data.domain[lat_attr]
        lon_attr = data.domain[lon_attr]

        fit_bounds = (self._domain != data.domain or
                      self.lat_attr is not lat_attr or
                      self.lon_attr is not lon_attr)
        self.lat_attr = lat_attr
        self.lon_attr = lon_attr
        self._domain = data.domain

        self._latlon_data = np.array([
            self.data.get_column_view(self.lat_attr)[0],
            self.data.get_column_view(self.lon_attr)[0]],
            dtype=float, order='F').T

        self._recompute_jittering_offsets()

        # Lat and/or Long is all-NaN. Clear the image and warn.
        if np.isnan(self._latlon_data).all(axis=0).any():
            self._owwidget.Warning.all_nan_slice()
            self.redraw_markers_overlay_image(new_image=True)
            return

        if fit_bounds:
            if self.isVisible():
                QTimer.singleShot(1, self.fit_to_bounds)
            else:
                self._should_fit_bounds = True
        else:
            self.redraw_markers_overlay_image(new_image=True)

    def showEvent(self, event):
        super().showEvent(event)
        if self._should_fit_bounds:
            QTimer.singleShot(500, self.fit_to_bounds)
            self._should_fit_bounds = False

    def fit_to_bounds(self, fly=True):
        if self.data is None:
            return
        lat_data, lon_data = self._latlon_data.T
        if np.isnan(lat_data).all() or np.isnan(lon_data).all():
            script = 'map.setView([0, 0], 0)'
        else:
            north, south = np.nanmax(lat_data), np.nanmin(lat_data)
            east, west = np.nanmin(lon_data), np.nanmax(lon_data)
            script = ('map.%sBounds([[%f, %f], [%f, %f]], {padding: [0,0], minZoom: 2, maxZoom: 13})' %
                      ('flyTo' if fly else 'fit', south, west, north, east))
        self.evalJS(script)
        # Sometimes on first data, it doesn't zoom in enough. So let do it
        # once more for good measure!
        self.evalJS(script)

    def selected_area(self, north, east, south, west):
        indices = np.array([])
        prev_selected_indices = self._selected_indices
        if self.data is not None and (north != south and east != west):
            lat, lon = self._latlon_data.T
            indices = ((lat <= north) & (lat >= south) &
                       (lon <= east) & (lon >= west))
            if self._selected_indices is not None:
                indices |= self._selected_indices
            self._selected_indices = indices
        else:
            self._selected_indices = None
        if np.any(self._selected_indices != prev_selected_indices):
            self.selectionChanged.emit(indices.nonzero()[0].tolist())
            self.redraw_markers_overlay_image(new_image=True)

    def set_map_provider(self, provider):
        self.evalJS('set_map_provider("{}");'.format(provider))

    def set_clustering(self, cluster_points):
        self.evalJS('''
            window.cluster_points = {};
            set_cluster_points();
        '''.format(int(cluster_points)))

    def _recompute_jittering_offsets(self):
        if not self._jittering:
            self._jittering_offsets = None
        elif self.data:
            # Calculate offsets randomly distributed within a circle
            screen_size = max(100, min(qApp.desktop().screenGeometry().width(),
                                       qApp.desktop().screenGeometry().height()))
            n = len(self.data)
            r = np.random.random(n)
            theta = np.random.uniform(0, 2*np.pi, n)
            xy_offsets = screen_size * self._jittering * np.c_[r * np.cos(theta),
                                                               r * np.sin(theta)]
            self._jittering_offsets = xy_offsets

    def set_jittering(self, jittering):
        """ In percent, i.e. jittering=3 means 3% of screen height and width """
        self._jittering = jittering / 100
        self._recompute_jittering_offsets()
        self.redraw_markers_overlay_image(new_image=True)

    @staticmethod
    def _legend_values(variable, values):
        strs = [variable.repr_val(val) for val in values]
        if any(len(val) > 10 for val in strs):
            if isinstance(variable, TimeVariable):
                strs = [s.replace(' ', '<br>') for s in strs]
            elif variable.is_continuous:
                strs = ['{:.4e}'.format(val) for val in values]
            elif variable.is_discrete:
                strs = [s if len(s) <= 12 else (s[:8] + '…' + s[-3:])
                        for s in strs]
        return strs

    def set_marker_color(self, attr, update=True):
        try:
            self._color_attr = variable = self.data.domain[attr]
            if len(self.data) == 0:
                raise Exception
        except Exception:
            self._color_attr = None
            self._legend_colors = []
        else:
            if variable.is_continuous:
                self._raw_color_values = values = self.data.get_column_view(variable)[0].astype(float)
                self._scaled_color_values = scale(values)
                self._colorgen = ContinuousPaletteGenerator(*variable.colors)

                colors = [color_to_hex(i) for i in variable.colors[:2]]
                if variable.colors[2]:  # pass through black
                    colors.insert(1, 'black')

                min = np.nanmin(values)
                self._legend_colors = (
                    ['c', self._legend_values(variable, [min, np.nanmax(values)]), colors]
                    if not np.isnan(min) else [])
            elif variable.is_discrete:
                _values = np.asarray(self.data.domain[attr].values)
                __values = self.data.get_column_view(variable)[0].astype(np.uint16)
                self._raw_color_values = _values[__values]  # The joke's on you
                self._scaled_color_values = __values
                self._colorgen = ColorPaletteGenerator(len(variable.colors), variable.colors)
                self._legend_colors = ['d',
                                       self._legend_values(variable, range(len(_values))),
                                       list(_values),
                                       [color_to_hex(self._colorgen.getRGB(i))
                                        for i in range(len(_values))]]
        finally:
            if update:
                self.redraw_markers_overlay_image(new_image=True)

    def set_marker_label(self, attr, update=True):
        try:
            self._label_attr = variable = self.data.domain[attr]
            if len(self.data) == 0:
                raise Exception
        except Exception:
            self._label_attr = None
        else:
            if variable.is_continuous or variable.is_string:
                self._label_values = self.data.get_column_view(variable)[0]
            elif variable.is_discrete:
                _values = np.asarray(self.data.domain[attr].values)
                __values = self.data.get_column_view(variable)[0].astype(np.uint16)
                self._label_values = _values[__values]  # The design had lead to poor code for ages
        finally:
            if update:
                self.redraw_markers_overlay_image(new_image=True)

    def set_marker_shape(self, attr, update=True):
        try:
            self._shape_attr = variable = self.data.domain[attr]
            if len(self.data) == 0:
                raise Exception
        except Exception:
            self._shape_attr = None
            self._legend_shapes = []
        else:
            assert variable.is_discrete
            _values = np.asarray(self.data.domain[attr].values)
            self._shape_values = __values = self.data.get_column_view(variable)[0].astype(np.uint16)
            self._raw_shape_values = _values[__values]
            self._legend_shapes = [self._legend_values(variable, range(len(_values))),
                                   list(_values)]
        finally:
            if update:
                self.redraw_markers_overlay_image(new_image=True)

    def set_marker_size(self, attr, update=True):
        try:
            self._size_attr = variable = self.data.domain[attr]
            if len(self.data) == 0:
                raise Exception
        except Exception:
            self._size_attr = None
            self._legend_sizes = []
        else:
            assert variable.is_continuous
            self._raw_sizes = values = self.data.get_column_view(variable)[0].astype(float)
            # Note, [5, 60] is also hardcoded in legend-size-indicator.svg
            self._sizes = scale(values, 5, 60).astype(np.uint8)
            min = np.nanmin(values)
            self._legend_sizes = self._legend_values(variable,
                                                     [min, np.nanmax(values)]) if not np.isnan(min) else []
        finally:
            if update:
                self.redraw_markers_overlay_image(new_image=True)

    def set_marker_size_coefficient(self, size):
        self._size_coef = size / 100
        self.evalJS('''set_marker_size_coefficient({});'''.format(size / 100))
        if not self.is_js_path:
            self.redraw_markers_overlay_image(new_image=True)

    def set_marker_opacity(self, opacity):
        self._opacity = 255 * opacity // 100
        self.evalJS('''set_marker_opacity({});'''.format(opacity / 100))
        if not self.is_js_path:
            self.redraw_markers_overlay_image(new_image=True)

    def set_model(self, model):
        self.model = model
        self.evalJS('clear_heatmap()' if model is None else 'reset_heatmap()')

    def recompute_heatmap(self, points):
        if self.model is None or self.data is None:
            self.exposeObject('model_predictions', {})
            self.evalJS('draw_heatmap()')
            return

        latlons = np.array(points)
        table = Table(Domain([self.lat_attr, self.lon_attr]), latlons)
        try:
            predictions = self.model(table)
        except Exception as e:
            self._owwidget.Error.model_error(e)
            return
        else:
            self._owwidget.Error.model_error.clear()

        class_var = self.model.domain.class_var
        is_regression = class_var.is_continuous
        if is_regression:
            predictions = scale(np.round(predictions, 7))  # Avoid small errors
            kwargs = dict(
                extrema=self._legend_values(class_var, [np.nanmin(predictions),
                                                        np.nanmax(predictions)]))
        else:
            colorgen = ColorPaletteGenerator(len(class_var.values), class_var.colors)
            predictions = colorgen.getRGB(predictions)
            kwargs = dict(
                legend_labels=self._legend_values(class_var, range(len(class_var.values))),
                full_labels=list(class_var.values),
                colors=[color_to_hex(colorgen.getRGB(i))
                        for i in range(len(class_var.values))])
        self.exposeObject('model_predictions', dict(data=predictions, **kwargs))
        self.evalJS('draw_heatmap()')

    def _update_legend(self, is_js_path=False):
        self.evalJS('''
            window.legend_colors = %s;
            window.legend_shapes = %s;
            window.legend_sizes  = %s;
            legendControl.remove();
            legendControl.addTo(map);
        ''' % (self._legend_colors,
               self._legend_shapes if is_js_path else [],
               self._legend_sizes))

    def _update_js_markers(self, visible, in_subset):
        self._visible = visible
        latlon = self._latlon_data
        self.exposeObject('latlon_data', dict(data=latlon[visible]))
        self.exposeObject('jittering_offsets',
                          self._jittering_offsets[visible] if self._jittering_offsets is not None else [])
        self.exposeObject('selected_markers', dict(data=(self._selected_indices[visible]
                                                         if self._selected_indices is not None else 0)))
        self.exposeObject('in_subset', in_subset.astype(np.int8))
        if not self._color_attr:
            self.exposeObject('color_attr', dict())
        else:
            colors = [color_to_hex(rgb)
                      for rgb in self._colorgen.getRGB(self._scaled_color_values[visible])]
            self.exposeObject('color_attr',
                              dict(name=str(self._color_attr), values=colors,
                                   raw_values=self._raw_color_values[visible]))
        if not self._label_attr:
            self.exposeObject('label_attr', dict())
        else:
            self.exposeObject('label_attr',
                              dict(name=str(self._label_attr),
                                   values=self._label_values[visible]))
        if not self._shape_attr:
            self.exposeObject('shape_attr', dict())
        else:
            self.exposeObject('shape_attr',
                              dict(name=str(self._shape_attr),
                                   values=self._shape_values[visible],
                                   raw_values=self._raw_shape_values[visible]))
        if not self._size_attr:
            self.exposeObject('size_attr', dict())
        else:
            self.exposeObject('size_attr',
                              dict(name=str(self._size_attr),
                                   values=self._sizes[visible],
                                   raw_values=self._raw_sizes[visible]))
        self.evalJS('''
            window.latlon_data = latlon_data.data;
            window.selected_markers = selected_markers.data;
            add_markers(latlon_data);
        ''')

    class Projection:
        """This should somewhat model Leaflet's Web Mercator (EPSG:3857).

        Reverse-engineered from L.Map.latlngToContainerPoint().
        """
        @staticmethod
        def latlon_to_easting_northing(lat, lon):
            R = 6378137
            MAX_LATITUDE = 85.0511287798
            DEG = np.pi / 180

            lat = np.clip(lat, -MAX_LATITUDE, MAX_LATITUDE)
            sin = np.sin(DEG * lat)
            x = R * DEG * lon
            y = R / 2 * np.log((1 + sin) / (1 - sin))
            return x, y

        @staticmethod
        def easting_northing_to_pixel(x, y, zoom_level, pixel_origin, map_pane_pos):
            R = 6378137
            PROJ_SCALE = .5 / (np.pi * R)

            zoom_scale = 256 * (2 ** zoom_level)
            x = (zoom_scale * (PROJ_SCALE * x + .5)).round() + (map_pane_pos[0] - pixel_origin[0])
            y = (zoom_scale * (-PROJ_SCALE * y + .5)).round() + (map_pane_pos[1] - pixel_origin[1])
            return x, y

    N_POINTS_PER_ITER = 666

    def redraw_markers_overlay_image(self, *args, new_image=False):
        if not args and not self._drawing_args or self.data is None:
            return

        if args:
            self._drawing_args = args
        north, east, south, west, width, height, zoom, origin, map_pane_pos = self._drawing_args

        lat, lon = self._latlon_data.T
        visible = ((lat <= north) & (lat >= south) &
                   (lon <= east) & (lon >= west)).nonzero()[0]
        in_subset = (np.in1d(self.data.ids, self._subset_ids)
                     if self._subset_ids.size else
                     np.tile(True, len(lon)))

        is_js_path = self.is_js_path = len(visible) < self.N_POINTS_PER_ITER

        self._update_legend(is_js_path)

        np.random.shuffle(visible)
        # Sort points in subset to be painted last
        visible = visible[np.lexsort((in_subset[visible],))]

        if is_js_path:
            self.evalJS('clear_markers_overlay_image()')
            self._update_js_markers(visible, in_subset[visible])
            self._owwidget.disable_some_controls(False)
            return

        self.evalJS('clear_markers_js();')
        self._owwidget.disable_some_controls(True)

        selected = (self._selected_indices
                    if self._selected_indices is not None else
                    np.zeros(len(lat), dtype=bool))
        cur = 0

        im = QImage(self._overlay_image_path)
        if im.isNull() or self._prev_origin != origin or new_image:
            im = QImage(width, height, QImage.Format_ARGB32)
            im.fill(Qt.transparent)
        else:
            dx, dy = self._prev_map_pane_pos - map_pane_pos
            im = im.copy(dx, dy, width, height)
        self._prev_map_pane_pos = np.array(map_pane_pos)
        self._prev_origin = origin

        painter = QPainter(im)
        painter.setRenderHint(QPainter.Antialiasing, True)
        self.evalJS('clear_markers_overlay_image(); markersImageLayer.setBounds(map.getBounds());0')

        self._image_token = image_token = np.random.random()

        n_iters = np.ceil(len(visible) / self.N_POINTS_PER_ITER)

        def add_points():
            nonlocal cur, image_token
            if image_token != self._image_token:
                return
            batch = visible[cur:cur + self.N_POINTS_PER_ITER]

            batch_lat = lat[batch]
            batch_lon = lon[batch]

            x, y = self.Projection.latlon_to_easting_northing(batch_lat, batch_lon)
            x, y = self.Projection.easting_northing_to_pixel(x, y, zoom, origin, map_pane_pos)

            if self._jittering:
                dx, dy = self._jittering_offsets[batch].T
                x, y = x + dx, y + dy

            colors = (self._colorgen.getRGB(self._scaled_color_values[batch]).tolist()
                      if self._color_attr else
                      repeat((0xff, 0, 0)))
            sizes = self._size_coef * \
                (self._sizes[batch] if self._size_attr else np.tile(10, len(batch)))

            opacity_subset, opacity_rest = self._opacity, int(.8 * self._opacity)
            for x, y, is_selected, size, color, _in_subset in \
                    zip(x, y, selected[batch], sizes, colors, in_subset[batch]):

                pensize2, selpensize2 = (.35, 1.5) if size >= 5 else (.15, .7)
                pensize2 *= self._size_coef
                selpensize2 *= self._size_coef

                size2 = size / 2
                if is_selected:
                    painter.setPen(QPen(QBrush(Qt.green), 2 * selpensize2))
                    painter.drawEllipse(x - size2 - selpensize2,
                                        y - size2 - selpensize2,
                                        size + selpensize2,
                                        size + selpensize2)
                color = QColor(*color)
                if _in_subset:
                    color.setAlpha(opacity_subset)
                    painter.setBrush(QBrush(color))
                    painter.setPen(QPen(QBrush(color.darker(180)), 2 * pensize2))
                else:
                    color.setAlpha(opacity_rest)
                    painter.setBrush(Qt.NoBrush)
                    painter.setPen(QPen(QBrush(color.lighter(120)), 2 * pensize2))

                painter.drawEllipse(x - size2 - pensize2,
                                    y - size2 - pensize2,
                                    size + pensize2,
                                    size + pensize2)

            im.save(self._overlay_image_path, 'PNG')
            self.evalJS('markersImageLayer.setUrl("{}#{}"); 0;'
                        .format(self.toFileURL(self._overlay_image_path),
                                np.random.random()))

            cur += self.N_POINTS_PER_ITER
            if cur < len(visible):
                QTimer.singleShot(10, add_points)
                self._owwidget.progressBarAdvance(100 / n_iters, None)
            else:
                self._owwidget.progressBarFinished(None)
                self._image_token = None

        self._owwidget.progressBarFinished(None)
        self._owwidget.progressBarInit(None)
        QTimer.singleShot(10, add_points)

    def set_subset_ids(self, ids):
        self._subset_ids = ids
        self.redraw_markers_overlay_image(new_image=True)

    def toggle_legend(self, visible):
        self.evalJS('''
            $(".legend").{0}();
            window.legend_hidden = "{0}";
        '''.format('show' if visible else 'hide'))


class OWMap(widget.OWWidget):
    name = 'Geo Map'
    description = 'Show data points on a world map.'
    icon = "icons/GeoMap.svg"
    priority = 100

    class Inputs:
        data = Input("Data", Table, default=True)
        data_subset = Input("Data Subset", Table)
        learner = Input("Learner", Learner)

    class Outputs:
        selected_data = Output("Selected Data", Table, default=True)
        annotated_data = Output(ANNOTATED_DATA_SIGNAL_NAME, Table)

    replaces = [
        "Orange.widgets.visualize.owmap.OWMap",
    ]

    settingsHandler = settings.DomainContextHandler()

    want_main_area = True

    autocommit = settings.Setting(True)
    tile_provider = settings.Setting('Black and white')
    lat_attr = settings.ContextSetting('')
    lon_attr = settings.ContextSetting('')
    class_attr = settings.ContextSetting('(None)')
    color_attr = settings.ContextSetting('')
    label_attr = settings.ContextSetting('')
    shape_attr = settings.ContextSetting('')
    size_attr = settings.ContextSetting('')
    opacity = settings.Setting(100)
    zoom = settings.Setting(100)
    jittering = settings.Setting(0)
    cluster_points = settings.Setting(False)
    show_legend = settings.Setting(True)

    TILE_PROVIDERS = OrderedDict((
        ('Black and white', 'OpenStreetMap.BlackAndWhite'),
        ('OpenStreetMap', 'OpenStreetMap.Mapnik'),
        ('Topographic', 'Thunderforest.OpenCycleMap'),
        ('Topographic 2', 'Thunderforest.Outdoors'),
        ('Satellite', 'Esri.WorldImagery'),
        ('Print', 'Stamen.TonerLite'),
        ('Dark', 'CartoDB.DarkMatter'),
        ('Watercolor', 'Stamen.Watercolor'),
    ))

    class Error(widget.OWWidget.Error):
        model_error = widget.Msg("Error predicting: {}")
        learner_error = widget.Msg("Error modelling: {}")

    class Warning(widget.OWWidget.Warning):
        all_nan_slice = widget.Msg('Latitude and/or longitude has no defined values (is all-NaN)')

    UserAdviceMessages = [
        widget.Message(
            'Select markers by holding <b><kbd>Shift</kbd></b> key and dragging '
            'a rectangle around them. Clear the selection by clicking anywhere.',
            'shift-selection')
    ]

    graph_name = "map"

    def __init__(self):
        super().__init__()
        self.map = map = LeafletMap(self)  # type: LeafletMap
        self.mainArea.layout().addWidget(map)
        self.selection = None
        self.data = None
        self.learner = None

        def selectionChanged(indices):
            self.selection = self.data[indices] if self.data is not None and indices else None
            self._indices = indices
            self.commit()

        map.selectionChanged.connect(selectionChanged)

        def _set_map_provider():
            map.set_map_provider(self.TILE_PROVIDERS[self.tile_provider])

        box = gui.vBox(self.controlArea, 'Map')
        gui.comboBox(box, self, 'tile_provider',
                     orientation=Qt.Horizontal,
                     label='Map:',
                     items=tuple(self.TILE_PROVIDERS.keys()),
                     sendSelectedValue=True,
                     callback=_set_map_provider)

        self._latlon_model = DomainModel(
            parent=self, valid_types=ContinuousVariable)
        self._class_model = DomainModel(
            parent=self, placeholder='(None)', valid_types=DomainModel.PRIMITIVE)
        self._color_model = DomainModel(
            parent=self, placeholder='(Same color)', valid_types=DomainModel.PRIMITIVE)
        self._shape_model = DomainModel(
            parent=self, placeholder='(Same shape)', valid_types=DiscreteVariable)
        self._size_model = DomainModel(
            parent=self, placeholder='(Same size)', valid_types=ContinuousVariable)
        self._label_model = DomainModel(
            parent=self, placeholder='(No labels)')

        def _set_lat_long():
            self.map.set_data(self.data, self.lat_attr, self.lon_attr)
            self.train_model()

        self._combo_lat = combo = gui.comboBox(
            box, self, 'lat_attr', orientation=Qt.Horizontal,
            label='Latitude:', sendSelectedValue=True, callback=_set_lat_long)
        combo.setModel(self._latlon_model)
        self._combo_lon = combo = gui.comboBox(
            box, self, 'lon_attr', orientation=Qt.Horizontal,
            label='Longitude:', sendSelectedValue=True, callback=_set_lat_long)
        combo.setModel(self._latlon_model)

        def _toggle_legend():
            self.map.toggle_legend(self.show_legend)

        gui.checkBox(box, self, 'show_legend', label='Show legend',
                     callback=_toggle_legend)

        box = gui.vBox(self.controlArea, 'Overlay')
        self._combo_class = combo = gui.comboBox(
            box, self, 'class_attr', orientation=Qt.Horizontal,
            label='Target:', sendSelectedValue=True, callback=self.train_model
        )
        self.controls.class_attr.setModel(self._class_model)
        self.set_learner(self.learner)

        box = gui.vBox(self.controlArea, 'Points')
        self._combo_color = combo = gui.comboBox(
            box, self, 'color_attr',
            orientation=Qt.Horizontal,
            label='Color:',
            sendSelectedValue=True,
            callback=lambda: self.map.set_marker_color(self.color_attr))
        combo.setModel(self._color_model)
        self._combo_label = combo = gui.comboBox(
            box, self, 'label_attr',
            orientation=Qt.Horizontal,
            label='Label:',
            sendSelectedValue=True,
            callback=lambda: self.map.set_marker_label(self.label_attr))
        combo.setModel(self._label_model)
        self._combo_shape = combo = gui.comboBox(
            box, self, 'shape_attr',
            orientation=Qt.Horizontal,
            label='Shape:',
            sendSelectedValue=True,
            callback=lambda: self.map.set_marker_shape(self.shape_attr))
        combo.setModel(self._shape_model)
        self._combo_size = combo = gui.comboBox(
            box, self, 'size_attr',
            orientation=Qt.Horizontal,
            label='Size:',
            sendSelectedValue=True,
            callback=lambda: self.map.set_marker_size(self.size_attr))
        combo.setModel(self._size_model)

        def _set_opacity():
            map.set_marker_opacity(self.opacity)

        def _set_zoom():
            map.set_marker_size_coefficient(self.zoom)

        def _set_jittering():
            map.set_jittering(self.jittering)

        def _set_clustering():
            map.set_clustering(self.cluster_points)

        self._opacity_slider = gui.hSlider(
            box, self, 'opacity', None, 1, 100, 5,
            label='Opacity:', labelFormat=' %d%%',
            callback=_set_opacity)
        self._zoom_slider = gui.valueSlider(
            box, self, 'zoom', None, values=(20, 50, 100, 200, 300, 400, 500, 700, 1000),
            label='Symbol size:', labelFormat=' %d%%',
            callback=_set_zoom)
        self._jittering = gui.valueSlider(
            box, self, 'jittering', label='Jittering:', values=(0, .5, 1, 2, 5),
            labelFormat=' %.1f%%', ticks=True,
            callback=_set_jittering)
        self._clustering_check = gui.checkBox(
            box, self, 'cluster_points', label='Cluster points',
            callback=_set_clustering)

        gui.rubber(self.controlArea)
        gui.auto_commit(self.controlArea, self, 'autocommit', 'Send Selection')

        QTimer.singleShot(0, _set_map_provider)
        QTimer.singleShot(0, _toggle_legend)
        QTimer.singleShot(0, _set_opacity)
        QTimer.singleShot(0, _set_zoom)
        QTimer.singleShot(0, _set_jittering)
        QTimer.singleShot(0, _set_clustering)

    autocommit = settings.Setting(True)

    def __del__(self):
        self.progressBarFinished(None)
        self.map = None

    def commit(self):
        self.Outputs.selected_data.send(self.selection)
        self.Outputs.annotated_data.send(create_annotated_table(self.data, self._indices))

    @Inputs.data
    def set_data(self, data):
        self.data = data

        self.closeContext()

        if data is None or not len(data):
            return self.clear()

        domain = data is not None and data.domain
        for model in (self._latlon_model,
                      self._class_model,
                      self._color_model,
                      self._shape_model,
                      self._size_model,
                      self._label_model):
            model.set_domain(domain)

        lat, lon = find_lat_lon(data)
        if lat or lon:
            self._combo_lat.setCurrentIndex(-1 if lat is None else self._latlon_model.indexOf(lat))
            self._combo_lon.setCurrentIndex(-1 if lat is None else self._latlon_model.indexOf(lon))
            self.lat_attr = lat.name
            self.lon_attr = lon.name

        if data.domain.class_var:
            self.color_attr = data.domain.class_var.name
        elif len(self._color_model):
            self._combo_color.setCurrentIndex(0)
        if len(self._shape_model):
            self._combo_shape.setCurrentIndex(0)
        if len(self._size_model):
            self._combo_size.setCurrentIndex(0)
        if len(self._label_model):
            self._combo_label.setCurrentIndex(0)
        if len(self._class_model):
            self._combo_class.setCurrentIndex(0)

        self.openContext(data)

        self.map.set_data(self.data, self.lat_attr, self.lon_attr)
        self.map.set_marker_color(self.color_attr, update=False)
        self.map.set_marker_label(self.label_attr, update=False)
        self.map.set_marker_shape(self.shape_attr, update=False)
        self.map.set_marker_size(self.size_attr, update=True)

    @Inputs.data_subset
    def set_subset(self, subset):
        self.map.set_subset_ids(subset.ids if subset is not None else np.array([]))

    def handleNewSignals(self):
        super().handleNewSignals()
        self.train_model()

    @Inputs.learner
    def set_learner(self, learner):
        self.learner = learner
        self.controls.class_attr.setEnabled(learner is not None)
        self.controls.class_attr.setToolTip(
            'Needs a Learner input for modelling.' if learner is None else '')

    def train_model(self):
        model = None
        self.Error.clear()
        if self.data and self.learner and self.class_attr != '(None)':
            domain = self.data.domain
            if self.lat_attr and self.lon_attr and self.class_attr in domain:
                domain = Domain([domain[self.lat_attr], domain[self.lon_attr]],
                                [domain[self.class_attr]])  # I am retarded
                train = Table.from_table(domain, self.data)
                try:
                    model = self.learner(train)
                except Exception as e:
                    self.Error.learner_error(e)
        self.map.set_model(model)

    def disable_some_controls(self, disabled):
        tooltip = (
            "Available when the zoom is close enough to have "
            "<{} points in the viewport.".format(self.map.N_POINTS_PER_ITER)
            if disabled else '')
        for widget in (self._combo_label,
                       self._combo_shape,
                       self._clustering_check):
            widget.setDisabled(disabled)
            widget.setToolTip(tooltip)

    def clear(self):
        self.map.set_data(None, '', '')
        for model in (self._latlon_model,
                      self._class_model,
                      self._color_model,
                      self._shape_model,
                      self._size_model,
                      self._label_model):
            model.set_domain(None)
        self.lat_attr = self.lon_attr = self.class_attr = self.color_attr = \
        self.label_attr = self.shape_attr = self.size_attr = None


def main():
    from AnyQt.QtWidgets import QApplication
    from Orange.modelling import KNNLearner as Learner
    a = QApplication([])

    ow = OWMap()
    ow.show()
    ow.raise_()
    data = Table("India_census_district_population")
    ow.set_data(data)

    QTimer.singleShot(10, lambda: ow.set_learner(Learner()))

    ow.handleNewSignals()
    a.exec()
    ow.saveSettings()

if __name__ == "__main__":
    main()
